@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2006-2012 Con Kolivas
+   Copyright (C) 2006-2015 Con Kolivas
    Copyright (C) 2011 Peter Hyman
    Copyright (C) 1998-2003 Andrew Tridgell
 
@@ -47,6 +47,7 @@
 #endif
 #include <math.h>
 #include <utime.h>
+#include <inttypes.h>
 
 #include "md5.h"
 #include "rzip.h"
@@ -104,7 +105,7 @@ i64 get_ram(rzip_control *control)
 	if(!(meminfo = fopen("/proc/meminfo", "r")))
 		fatal_return(("fopen\n"), -1);
 
-	while(!feof(meminfo) && !fscanf(meminfo, "MemTotal: %lld kB", &ramsize)) {
+	while(!feof(meminfo) && !fscanf(meminfo, "MemTotal: %"PRId64" kB", &ramsize)) {
 		if (unlikely(fgets(aux, sizeof(aux), meminfo) == NULL)) {
 			fclose(meminfo);
 			fatal_return(("Failed to fgets in get_ram\n"), -1);
@@ -136,12 +137,9 @@ i64 nloops(i64 seconds, uchar *b1, uchar *b2)
 
 bool write_magic(rzip_control *control)
 {
-	char magic[MAGIC_LEN];
-
-	memset(magic, 0, MAGIC_LEN);
-	strcpy(magic, "LRZI");
-	magic[4] = LRZIP_MAJOR_VERSION;
-	magic[5] = LRZIP_MINOR_VERSION;
+	char magic[MAGIC_LEN] = { 
+		'L', 'R', 'Z', 'I', LRZIP_MAJOR_VERSION, LRZIP_MINOR_VERSION 
+	};
 
 	/* File size is stored as zero for streaming STDOUT blocks when the
 	 * file size is unknown. In encrypted files, the size is left unknown
@@ -309,17 +307,14 @@ int open_tmpoutfile(rzip_control *control)
 			fatal_return(("Failed to allocate outfile name\n"), -1);
 		strcpy(control->outfile, control->tmpdir);
 		strcat(control->outfile, "lrzipout.XXXXXX");
-	} else {
-		control->outfile = realloc(NULL, 16);
-		if (unlikely(!control->outfile))
-			fatal_return(("Failed to allocate outfile name\n"), -1);
-		strcpy(control->outfile, "lrzipout.XXXXXX");
 	}
 
 	fd_out = mkstemp(control->outfile);
-	if (unlikely(fd_out == -1))
-		fatal_return(("Failed to create out tmpfile: %s\n", control->outfile), -1);
-	register_outfile(control, control->outfile, TEST_ONLY || STDOUT || !KEEP_BROKEN);
+	if (fd_out == -1) {
+		print_verbose("WARNING: Failed to create out tmpfile: %s , will fail if cannot perform entirely in ram\n",
+			      control->outfile, DECOMPRESS ? "de" : "");
+	} else
+		register_outfile(control, control->outfile, TEST_ONLY || STDOUT || !KEEP_BROKEN);
 	return fd_out;
 }
 
@@ -355,7 +350,7 @@ bool write_fdout(rzip_control *control, void *buf, i64 len)
 		ret = MIN(len, one_g);
 		ret = write(control->fd_out, offset_buf, (size_t)ret);
 		if (unlikely(ret <= 0))
-			fatal_return(("Failed to write to fd_out in write_fdout\n"), false);;
+			fatal_return(("Failed to write to fd_out in write_fdout\n"), false);
 		len -= ret;
 		offset_buf += ret;
 	}
@@ -385,6 +380,8 @@ bool dump_tmpoutfile(rzip_control *control, int fd_out)
 	FILE *tmpoutfp;
 	int tmpchar;
 
+	if (unlikely(fd_out == -1))
+		fatal_return(("Failed: No temporary outfile created, unable to do in ram\n"), false);
 	/* flush anything not yet in the temporary file */
 	fsync(fd_out);
 	tmpoutfp = fdopen(fd_out, "r");
@@ -507,10 +504,15 @@ static bool open_tmpoutbuf(rzip_control *control)
 	return true;
 }
 
+/* We've decided to use a temporary output file instead of trying to store
+ * all the output buffer in ram so we can free up the ram and increase the
+ * maximum sizes of ram we can allocate */
 void close_tmpoutbuf(rzip_control *control)
 {
 	control->flags &= ~FLAG_TMP_OUTBUF;
 	free(control->tmp_outbuf);
+	if (!BITS32)
+		control->usable_ram = control->maxram += control->ramsize / 18;
 }
 
 static bool open_tmpinbuf(rzip_control *control)
@@ -537,10 +539,13 @@ bool clear_tmpinfile(rzip_control *control)
 	return true;
 }
 
+/* As per temporary output file but for input file */
 void close_tmpinbuf(rzip_control *control)
 {
 	control->flags &= ~FLAG_TMP_INBUF;
 	free(control->tmp_inbuf);
+	if (!BITS32)
+		control->usable_ram = control->maxram += control->ramsize / 18;
 }
 
 static int get_pass(rzip_control *control, char *s)
@@ -737,14 +742,16 @@ bool decompress_file(rzip_control *control)
 				return false;
 	} else {
 		fd_out = open_tmpoutfile(control);
-		if (unlikely(fd_out == -1))
-			fatal_return(("Failed to create %s\n", control->outfile), false);
-		fd_hist = open(control->outfile, O_RDONLY);
-		if (unlikely(fd_hist == -1))
-			fatal_return(("Failed to open history file %s\n", control->outfile), false);
-		/* Unlink temporary file as soon as possible */
-		if (unlikely(unlink(control->outfile)))
-			fatal_return(("Failed to unlink tmpfile: %s\n", control->outfile), false);
+		if (fd_out == -1) {
+			fd_hist = -1;
+		} else {
+			fd_hist = open(control->outfile, O_RDONLY);
+			if (unlikely(fd_hist == -1))
+				fatal_return(("Failed to open history file %s\n", control->outfile), false);
+			/* Unlink temporary file as soon as possible */
+			if (unlikely(unlink(control->outfile)))
+				fatal_return(("Failed to unlink tmpfile: %s\n", control->outfile), false);
+		}
 	}
 
 	if (unlikely(!open_tmpoutbuf(control)))
@@ -754,7 +761,7 @@ bool decompress_file(rzip_control *control)
 		if (unlikely(!read_magic(control, fd_in, &expected_size)))
 			return false;
 
-	if (!STDOUT) {
+	if (!STDOUT && !TEST_ONLY) {
 		/* Check if there's enough free space on the device chosen to fit the
 		* decompressed file. */
 		if (unlikely(fstatvfs(fd_out, &fbuf)))
@@ -787,9 +794,10 @@ bool decompress_file(rzip_control *control)
 	if (unlikely(runzip_fd(control, fd_in, fd_out, fd_hist, expected_size) < 0))
 		return false;
 
-	if (STDOUT && !TMP_OUTBUF)
+	if (STDOUT && !TMP_OUTBUF) {
 		if (unlikely(!dump_tmpoutfile(control, fd_out)))
 			return false;
+	}
 
 	/* if we get here, no fatal_return(( errors during decompression */
 	print_progress("\r");
@@ -802,8 +810,10 @@ bool decompress_file(rzip_control *control)
 	else
 		print_progress("[OK]                                             \n");
 
-	if (unlikely(close(fd_hist) || close(fd_out)))
-		fatal_return(("Failed to close files\n"), false);
+	if (fd_out > 0) {
+		if (unlikely(close(fd_hist) || close(fd_out)))
+			fatal_return(("Failed to close files\n"), false);
+	}
 
 	if (unlikely(!STDIN && !STDOUT && !TEST_ONLY && !preserve_times(control, fd_in)))
 		return false;
@@ -870,7 +880,9 @@ static double percentage(i64 num, i64 den)
 
 	if (den < 100) {
 		d_num = num * 100;
-		d_den = den ? : 1;
+		d_den = den;
+		if (!d_den)
+			d_den = 1;
 	} else {
 		d_num = num;
 		d_den = den / 100;
@@ -968,7 +980,7 @@ next_chunk:
 	while (stream < NUM_STREAMS) {
 		int block = 1;
 
-		if (unlikely(lseek(fd_in, stream_head[stream] + ofs, SEEK_SET)) == -1)
+		if (unlikely(lseek(fd_in, stream_head[stream] + ofs, SEEK_SET) == -1))
 			fatal_goto(("Failed to seek to header data in get_fileinfo\n"), error);
 		if (unlikely(!get_header_info(control, fd_in, &ctype, &c_len, &u_len, &last_head, chunk_byte)))
 			return false;
@@ -981,7 +993,7 @@ next_chunk:
 
 			if (unlikely(last_head + ofs > infile_size))
 				failure_goto(("Offset greater than archive size, likely corrupted/truncated archive.\n"), error);
-			if (unlikely(head_off = lseek(fd_in, last_head + ofs, SEEK_SET)) == -1)
+			if (unlikely(head_off = lseek(fd_in, last_head + ofs, SEEK_SET) == -1))
 				fatal_goto(("Failed to seek to header data in get_fileinfo\n"), error);
 			if (unlikely(!get_header_info(control, fd_in, &ctype, &c_len, &u_len,
 					&last_head, chunk_byte)))
@@ -1037,8 +1049,6 @@ next_chunk:
 done:
 	if (unlikely(ofs > infile_size))
 		failure_goto(("Offset greater than archive size, likely corrupted/truncated archive.\n"), error);
-	if (chunk_total > expected_size)
-		expected_size = chunk_total;
 	print_verbose("Rzip compression: %.1f%% %lld / %lld\n",
 			percentage (utotal, expected_size),
 			utotal, expected_size);
@@ -1077,7 +1087,7 @@ done:
 		int i;
 
 		print_output("MD5 used for integrity testing\n");
-		if (unlikely(lseek(fd_in, -MD5_DIGEST_SIZE, SEEK_END)) == -1)
+		if (unlikely(lseek(fd_in, -MD5_DIGEST_SIZE, SEEK_END) == -1))
 			fatal_goto(("Failed to seek to md5 data in runzip_fd\n"), error);
 		if (unlikely(read(fd_in, md5_stored, MD5_DIGEST_SIZE) != MD5_DIGEST_SIZE))
 			fatal_goto(("Failed to read md5 data in runzip_fd\n"), error);
@@ -1220,7 +1230,7 @@ error:
 	return false;
 }
 
-bool initialize_control(rzip_control *control)
+bool initialise_control(rzip_control *control)
 {
 	struct timeval tv;
 	char *eptr; /* for environment */
@@ -1250,17 +1260,29 @@ bool initialize_control(rzip_control *control)
 	if (unlikely(!get_rand(control, control->salt + 2, 6)))
 		return false;
 
-	/* Get Temp Dir */
-	eptr = getenv("TMP");
-	if (eptr != NULL) {
-		size_t len = strlen(eptr);
-		control->tmpdir = malloc(len+2);
-		if (control->tmpdir == NULL)
-			fatal_return(("Failed to allocate for tmpdir\n"), false);
-		strcpy(control->tmpdir, eptr);
-		if (eptr[len - 2] != '/')
-			eptr[len - 2] = '/'; /* need a trailing slash */
-		eptr[len - 1] = 0;
+	/* Get Temp Dir. Try variations on canonical unix environment variable */
+	eptr = getenv("TMPDIR");
+	if (!eptr)
+		eptr = getenv("TMP");
+	if (!eptr)
+		eptr = getenv("TEMPDIR");
+	if (!eptr)
+		eptr = getenv("TEMP");
+	if (!eptr) {
+		eptr = malloc(3);
+		if ( eptr == NULL )
+			fatal_return(("Failed to allocate for eptr\n"), false);
+		strcpy(eptr,"./");
 	}
+	size_t len = strlen(eptr);
+
+	control->tmpdir = malloc(len+2);
+	if (control->tmpdir == NULL)
+		fatal_return(("Failed to allocate for tmpdir\n"), false);
+	strcpy(control->tmpdir, eptr);
+	if (control->tmpdir[len] != '/')
+		control->tmpdir[len] = '/'; /* need a trailing slash */
+	control->tmpdir[len+1] = '\0';
+
 	return true;
 }
