@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2006-2013 Con Kolivas
+   Copyright (C) 2006-2015 Con Kolivas
    Copyright (C) 1998 Andrew Tridgell
 
    Modified to use flat hash, memory limit and variable hash culling
@@ -73,7 +73,6 @@
  * even tags, then all tags divisible by four, etc.).  This ensures
  * that on average, all parts of the file are covered by the hash, if
  * sparsely. */
-typedef i64 tag;
 
 /* All zero means empty.  We might miss the first chunk this way. */
 struct hash_entry {
@@ -99,35 +98,7 @@ static struct level {
 	{ 64, 1, 128 },
 };
 
-struct rzip_state {
-	void *ss;
-	struct level *level;
-	tag hash_index[256];
-	struct hash_entry *hash_table;
-	char hash_bits;
-	i64 hash_count;
-	i64 hash_limit;
-	tag minimum_tag_mask;
-	i64 tag_clean_ptr;
-	i64 last_match;
-	i64 chunk_size;
-	i64 mmap_size;
-	char chunk_bytes;
-	uint32_t cksum;
-	int fd_in, fd_out;
-	char stdin_eof;
-	struct {
-		i64 inserts;
-		i64 literals;
-		i64 literal_bytes;
-		i64 matches;
-		i64 match_bytes;
-		i64 tag_hits;
-		i64 tag_misses;
-	} stats;
-};
-
-static bool remap_low_sb(rzip_control *control, struct sliding_buffer *sb)
+static void remap_low_sb(rzip_control *control, struct sliding_buffer *sb)
 {
 	i64 new_offset;
 
@@ -135,20 +106,19 @@ static bool remap_low_sb(rzip_control *control, struct sliding_buffer *sb)
 	round_to_page(&new_offset);
 	print_maxverbose("Sliding main buffer to offset %lld\n", new_offset);
 	if (unlikely(munmap(sb->buf_low, sb->size_low)))
-		fatal_return(("Failed to munmap in remap_low_sb\n"), false);
+		failure("Failed to munmap in remap_low_sb\n");
 	if (new_offset + sb->size_low > sb->orig_size)
 		sb->size_low = sb->orig_size - new_offset;
 	sb->offset_low = new_offset;
 	sb->buf_low = (uchar *)mmap(sb->buf_low, sb->size_low, PROT_READ, MAP_SHARED, sb->fd, sb->orig_offset + sb->offset_low);
 	if (unlikely(sb->buf_low == MAP_FAILED))
-		fatal_return(("Failed to re mmap in remap_low_sb\n"), false);
-	return true;
+		failure("Failed to re mmap in remap_low_sb\n");
 }
 
-static inline bool remap_high_sb(rzip_control *control, struct sliding_buffer *sb, i64 p)
+static inline void remap_high_sb(rzip_control *control, struct sliding_buffer *sb, i64 p)
 {
 	if (unlikely(munmap(sb->buf_high, sb->size_high)))
-		fatal_return(("Failed to munmap in remap_high_sb\n"), false);
+		failure("Failed to munmap in remap_high_sb\n");
 	sb->size_high = sb->high_length; /* In case we shrunk it when we hit the end of the file */
 	sb->offset_high = p;
 	/* Make sure offset is rounded to page size of total offset */
@@ -157,8 +127,7 @@ static inline bool remap_high_sb(rzip_control *control, struct sliding_buffer *s
 		sb->size_high = sb->orig_size - sb->offset_high;
 	sb->buf_high = (uchar *)mmap(sb->buf_high, sb->size_high, PROT_READ, MAP_SHARED, sb->fd, sb->orig_offset + sb->offset_high);
 	if (unlikely(sb->buf_high == MAP_FAILED))
-		fatal_return(("Failed to re mmap in remap_high_sb\n"), false);
-	return true;
+		failure("Failed to re mmap in remap_high_sb\n");
 }
 
 /* We use a "sliding mmap" to effectively read more than we can fit into the
@@ -182,17 +151,11 @@ static uchar *sliding_get_sb(rzip_control *control, i64 p)
 	if (p >= sbo && p < (sbo + sb->size_high))
 		return (sb->buf_high + (p - sbo));
 	/* p is not within the low or high buffer range */
-	if (unlikely(!remap_high_sb(control, &control->sb, p)))
-		return NULL;
+	remap_high_sb(control, &control->sb, p);
 	/* Use sb->offset_high directly since it will have changed */
 	return (sb->buf_high + (p - sb->offset_high));
 }
 
-static uchar *single_get_sb(__maybe_unused rzip_control *control, i64 p)
-{
-	return (control->sb.buf_low + p);
-}
-
 /* The length of continous range of the sliding buffer,
  * starting from the offset P.
  */
@@ -235,40 +198,31 @@ static void sliding_mcpy(rzip_control *control, unsigned char *buf, i64 offset,
 }
 
 /* All put_u8/u32/vchars go to stream 0 */
-static inline bool put_u8(rzip_control *control, void *ss, uchar b)
+static inline void put_u8(rzip_control *control, void *ss, uchar b)
 {
-	if (unlikely(write_stream(control, ss, 0, &b, 1)))
-		fatal_return(("Failed to put_u8\n"), false);
-	return true;
+	write_stream(control, ss, 0, &b, 1);
 }
 
-static inline bool put_u32(rzip_control *control, void *ss, uint32_t s)
+static inline void put_u32(rzip_control *control, void *ss, uint32_t s)
 {
 	s = htole32(s);
-	if (unlikely(write_stream(control, ss, 0, (uchar *)&s, 4)))
-		fatal_return(("Failed to put_u32\n"), false);
-	return true;
+	write_stream(control, ss, 0, (uchar *)&s, 4);
 }
 
 /* Put a variable length of bytes dependant on how big the chunk is */
-static bool put_vchars(rzip_control *control, void *ss, i64 s, i64 length)
+static void put_vchars(rzip_control *control, void *ss, i64 s, i64 length)
 {
 	s = htole64(s);
-	if (unlikely(write_stream(control, ss, 0, (uchar *)&s, length)))
-		fatal_return(("Failed to put_vchars\n"), false);
-	return true;
+	write_stream(control, ss, 0, (uchar *)&s, length);
 }
 
-static bool put_header(rzip_control *control, void *ss, uchar head, i64 len)
+static void put_header(rzip_control *control, void *ss, uchar head, i64 len)
 {
-	if (unlikely(!put_u8(control, ss, head)))
-		return false;
-	if (unlikely(!put_vchars(control, ss, len, 2)))
-		return false;
-	return true;
+	put_u8(control, ss, head);
+	put_vchars(control, ss, len, 2);
 }
 
-static inline bool put_match(rzip_control *control, struct rzip_state *st,
+static inline void put_match(rzip_control *control, struct rzip_state *st,
 			     i64 p, i64 offset, i64 len)
 {
 	do {
@@ -278,22 +232,18 @@ static inline bool put_match(rzip_control *control, struct rzip_state *st,
 			n = 0xFFFF;
 
 		ofs = (p - offset);
-		if (unlikely(!put_header(control, st->ss, 1, n)))
-			return false;
-		if (unlikely(!put_vchars(control, st->ss, ofs, st->chunk_bytes)))
-			return false;
-
+		put_header(control, st->ss, 1, n);
+		put_vchars(control, st->ss, ofs, st->chunk_bytes);
 		st->stats.matches++;
 		st->stats.match_bytes += n;
 		len -= n;
 		p += n;
 		offset += n;
 	} while (len);
-	return true;
 }
 
 /* write some data to a stream mmap encoded. Return -1 on failure */
-static inline int write_sbstream(rzip_control *control, void *ss, int stream,
+static inline void write_sbstream(rzip_control *control, void *ss, int stream,
 				 i64 p, i64 len)
 {
 	struct stream_info *sinfo = ss;
@@ -308,30 +258,27 @@ static inline int write_sbstream(rzip_control *control, void *ss, int stream,
 		len -= n;
 
 		if (sinfo->s[stream].buflen == sinfo->bufsize)
-			if (unlikely(!flush_buffer(control, sinfo, stream)))
-				return -1;
+			flush_buffer(control, sinfo, stream);
 	}
-	return 0;
 }
 
-static bool put_literal(rzip_control *control, struct rzip_state *st, i64 last, i64 p)
+static void put_literal(rzip_control *control, struct rzip_state *st, i64 last, i64 p)
 {
 	do {
 		i64 len = p - last;
+
 		if (len > 0xFFFF)
 			len = 0xFFFF;
 
 		st->stats.literals++;
 		st->stats.literal_bytes += len;
 
-		if (unlikely(!put_header(control, st->ss, 0, len)))
-			return false;
+		put_header(control, st->ss, 0, len);
 
-		if (unlikely(len && write_sbstream(control, st->ss, 1, last, len)))
-			fatal_return(("Failed to write_stream in put_literal\n"), false);
+		if (len)
+			write_sbstream(control, st->ss, 1, last, len);
 		last += len;
 	} while (p > last);
-	return true;
 }
 
 /* Could give false positive on offset 0.  Who cares. */
@@ -452,44 +399,63 @@ again:
 	goto again;
 }
 
-static inline void next_tag(rzip_control *control, struct rzip_state *st, i64 p, tag *t)
+static void single_next_tag(rzip_control *control, struct rzip_state *st, i64 p, tag *t)
+{
+	uchar u;
+
+	u = control->sb.buf_low[p - 1];
+	*t ^= st->hash_index[u];
+	u = control->sb.buf_low[p + MINIMUM_MATCH - 1];
+	*t ^= st->hash_index[u];
+}
+
+static void sliding_next_tag(rzip_control *control, struct rzip_state *st, i64 p, tag *t)
 {
 	uchar *u;
 
-	u = control->get_sb(control, p - 1);
+	u = sliding_get_sb(control, p - 1);
 	*t ^= st->hash_index[*u];
-	u = control->get_sb(control, p + MINIMUM_MATCH - 1);
+	u = sliding_get_sb(control, p + MINIMUM_MATCH - 1);
 	*t ^= st->hash_index[*u];
 }
 
-static inline tag full_tag(rzip_control *control, struct rzip_state *st, i64 p)
+static tag single_full_tag(rzip_control *control, struct rzip_state *st, i64 p)
+{
+	tag ret = 0;
+	int i;
+	uchar u;
+
+	for (i = 0; i < MINIMUM_MATCH; i++) {
+		u = control->sb.buf_low[p + i];
+		ret ^= st->hash_index[u];
+	}
+	return ret;
+}
+
+static tag sliding_full_tag(rzip_control *control, struct rzip_state *st, i64 p)
 {
 	tag ret = 0;
 	int i;
 	uchar *u;
 
 	for (i = 0; i < MINIMUM_MATCH; i++) {
-		u = control->get_sb(control, p + i);
-		if (unlikely(!u))
-			return -1;
+		u = sliding_get_sb(control, p + i);
 		ret ^= st->hash_index[*u];
 	}
 	return ret;
 }
 
-static inline i64
-match_len(rzip_control *control, struct rzip_state *st, i64 p0, i64 op,
-	  i64 end, i64 *rev)
+static i64
+single_match_len(rzip_control *control, struct rzip_state *st, i64 p0, i64 op,
+		 i64 end, i64 *rev)
 {
-	uchar *(*csb)(rzip_control *, i64);
 	i64 p, len;
 
 	if (op >= p0)
 		return 0;
 
 	p = p0;
-	csb = control->get_sb;
-	while (p < end && (*csb(control, p) == *csb(control, op))) {
+	while (p < end && control->sb.buf_low[p] == control->sb.buf_low[op]) {
 		p++;
 		op++;
 	}
@@ -499,7 +465,39 @@ match_len(rzip_control *control, struct rzip_state *st, i64 p0, i64 op,
 
 	end = MAX(0, st->last_match);
 
-	while (p > end && op > 0 && *csb(control, op - 1) == *csb(control, p - 1)) {
+	while (p > end && op > 0 && control->sb.buf_low[op - 1] == control->sb.buf_low[p - 1]) {
+		op--;
+		p--;
+	}
+
+	len += *rev = p0 - p;
+	if (len < MINIMUM_MATCH)
+		return 0;
+
+	return len;
+}
+
+static i64
+sliding_match_len(rzip_control *control, struct rzip_state *st, i64 p0, i64 op,
+		  i64 end, i64 *rev)
+{
+	i64 p, len;
+
+	if (op >= p0)
+		return 0;
+
+	p = p0;
+	while (p < end && *sliding_get_sb(control, p) == *sliding_get_sb(control, op)) {
+		p++;
+		op++;
+	}
+	len = p - p0;
+	p = p0;
+	op -= len;
+
+	end = MAX(0, st->last_match);
+
+	while (p > end && op > 0 && *sliding_get_sb(control, op - 1) == *sliding_get_sb(control, p - 1)) {
 		op--;
 		p--;
 	}
@@ -531,19 +529,17 @@ find_best_match(rzip_control *control, struct rzip_state *st, tag t, i64 p,
 		i64 mlen;
 
 		if (t == he->t) {
-			mlen = match_len(control, st, p, he->offset, end,
-					 &rev);
-
-			if (mlen)
+			mlen = control->match_len(control, st, p, he->offset, end,
+						  &rev);
+			if (mlen) {
+				if (mlen > length) {
+					length = mlen;
+					(*offset) = he->offset - rev;
+					(*reverse) = rev;
+				}
 				st->stats.tag_hits++;
-			else
+			} else
 				st->stats.tag_misses++;
-
-			if (mlen >= length) {
-				length = mlen;
-				(*offset) = he->offset - rev;
-				(*reverse) = rev;
-			}
 		}
 
 		h++;
@@ -573,8 +569,12 @@ static void show_distrib(rzip_control *control, struct rzip_state *st)
 	if (total != st->hash_count)
 		print_err("WARNING: hash_count says total %lld\n", st->hash_count);
 
-	print_output("%lld total hashes -- %lld in primary bucket (%-2.3f%%)\n", total, primary,
-		     primary * 100.0 / (total ? : 1));
+	if (!total)
+		print_output("0 total hashes\n");
+	else {
+		print_output("%lld total hashes -- %lld in primary bucket (%-2.3f%%)\n",
+			     total, primary, primary * 100.0 / total);
+	}
 }
 
 /* Perform all checksumming in a separate thread to speed up the hash search. */
@@ -588,7 +588,7 @@ static void *cksumthread(void *data)
 	if (!NO_MD5)
 		md5_process_bytes(control->checksum.buf, control->checksum.len, &control->ctx);
 	free(control->checksum.buf);
-	unlock_mutex(control, &control->cksumlock);
+	cksem_post(control, &control->cksumsem);
 	return NULL;
 }
 
@@ -599,21 +599,19 @@ static inline void cksum_update(rzip_control *control)
 	create_pthread(control, &thread, NULL, cksumthread, control);
 }
 
-static inline bool hash_search(rzip_control *control, struct rzip_state *st,
+static inline void hash_search(rzip_control *control, struct rzip_state *st,
 			       double pct_base, double pct_multiple)
 {
 	i64 cksum_limit = 0, p, end, cksum_chunks, cksum_remains, i;
+	tag t = 0, tag_mask = (1 << st->level->initial_freq) - 1;
 	struct sliding_buffer *sb = &control->sb;
 	int lastpct = 0, last_chunkpct = 0;
-	tag t = 0;
 	struct {
 		i64 p;
 		i64 ofs;
 		i64 len;
 	} current;
 
-	tag tag_mask = (1 << st->level->initial_freq) - 1;
-
 	if (st->hash_table)
 		memset(st->hash_table, 0, sizeof(st->hash_table[0]) * (1<<st->hash_bits));
 	else {
@@ -628,7 +626,7 @@ static inline bool hash_search(rzip_control *control, struct rzip_state *st,
 		st->hash_limit = (1 << st->hash_bits) / 3 * 2;
 		st->hash_table = calloc(sizeof(st->hash_table[0]), (1 << st->hash_bits));
 		if (unlikely(!st->hash_table))
-			fatal_return(("Failed to allocate hash table in hash_search\n"), false);
+			failure("Failed to allocate hash table in hash_search\n");
 	}
 
 	st->minimum_tag_mask = tag_mask;
@@ -643,11 +641,8 @@ static inline bool hash_search(rzip_control *control, struct rzip_state *st,
 	current.p = p;
 	current.ofs = 0;
 
-	if (likely(end > 0)) {
-		t = full_tag(control, st, p);
-		if (unlikely(t == -1))
-			return false;
-	}
+	if (likely(end > 0))
+		t = control->full_tag(control, st, p);
 
 	while (p < end) {
 		i64 reverse, mlen, offset;
@@ -656,12 +651,12 @@ static inline bool hash_search(rzip_control *control, struct rzip_state *st,
 		if (unlikely(sb->offset_search > sb->offset_low + sb->size_low))
 			remap_low_sb(control, &control->sb);
 
-		if (unlikely(p % 128 == 0)) {
-			int pct, chunk_pct;
+		if (unlikely(p % 128 == 0 && st->chunk_size)) {
+			i64 chunk_pct;
+			int pct;
 
-			pct = pct_base + (pct_multiple * (100.0 * p) /
-			      (st->chunk_size ? : 1));
-			chunk_pct = p / ((end / 100) ? : 1);
+			pct = pct_base + (pct_multiple * (100.0 * p) / st->chunk_size );
+			chunk_pct = p * 100 / end;
 			if (pct != lastpct || chunk_pct != last_chunkpct) {
 				if (!STDIN || st->stdin_eof)
 					print_progress("Total: %2d%%  ", pct);
@@ -674,7 +669,7 @@ static inline bool hash_search(rzip_control *control, struct rzip_state *st,
 			}
 		}
 
-		next_tag(control, st, p, &t);
+		control->next_tag(control, st, p, &t);
 
 		/* Don't look for a match if there are no tags with
 		   this number of bits in the hash table. */
@@ -702,16 +697,12 @@ static inline bool hash_search(rzip_control *control, struct rzip_state *st,
 		if ((current.len >= GREAT_MATCH || p >= current.p + MINIMUM_MATCH)
 		    && current.len >= MINIMUM_MATCH) {
 			if (st->last_match < current.p)
-				if (unlikely(!put_literal(control, st, st->last_match, current.p)))
-					return false;
-			if (unlikely(!put_match(control, st, current.p, current.ofs, current.len)))
-				return false;
+				put_literal(control, st, st->last_match, current.p);
+			put_match(control, st, current.p, current.ofs, current.len);
 			st->last_match = current.p + current.len;
 			current.p = p = st->last_match;
 			current.len = 0;
-			t = full_tag(control, st, p);
-			if (unlikely(t == -1))
-				return false;
+			t = control->full_tag(control, st, p);
 		}
 
 		if (p > cksum_limit) {
@@ -719,11 +710,11 @@ static inline bool hash_search(rzip_control *control, struct rzip_state *st,
 			 * cksumthread. This lock protects all the data in
 			 * control->checksum.
 			 */
-			lock_mutex(control, &control->cksumlock);
+			cksem_wait(control, &control->cksumsem);
 			control->checksum.len = MIN(st->chunk_size - p, control->page_size);
 			control->checksum.buf = malloc(control->checksum.len);
 			if (unlikely(!control->checksum.buf))
-				fatal_return(("Failed to malloc ckbuf in hash_search\n"), false);
+				failure("Failed to malloc ckbuf in hash_search\n");
 			control->do_mcpy(control, control->checksum.buf, cksum_limit, control->checksum.len);
 			control->checksum.cksum = &st->cksum;
 			cksum_update(control);
@@ -740,14 +731,14 @@ static inline bool hash_search(rzip_control *control, struct rzip_state *st,
 	if (st->chunk_size > cksum_limit) {
 		/* Compute checksum. If the entire chunk is longer than maxram,
 		 * do it "per-partes" */
-		lock_mutex(control, &control->cksumlock);
+		cksem_wait(control, &control->cksumsem);
 		control->checksum.len = st->chunk_size - cksum_limit;
 		cksum_chunks = control->checksum.len / control->maxram;
 		cksum_remains = control->checksum.len % control->maxram;
 
-		control->checksum.buf = malloc(control->checksum.len);
+		control->checksum.buf = malloc(control->maxram);
 		if (unlikely(!control->checksum.buf))
-			fatal_return(("Failed to malloc ckbuf in hash_search2\n"), false);
+			failure("Failed to malloc ckbuf in hash_search2\n");
 
 		for (i = 0; i < cksum_chunks; i++) {
 			control->do_mcpy(control, control->checksum.buf, cksum_limit, control->maxram);
@@ -762,15 +753,14 @@ static inline bool hash_search(rzip_control *control, struct rzip_state *st,
 		if (!NO_MD5)
 			md5_process_bytes(control->checksum.buf, cksum_remains, &control->ctx);
 		free(control->checksum.buf);
-		unlock_mutex(control, &control->cksumlock);
-	} else
-		wait_mutex(control, &control->cksumlock);
+		cksem_post(control, &control->cksumsem);
+	} else {
+		cksem_wait(control, &control->cksumsem);
+		cksem_post(control, &control->cksumsem);
+	}
 
-	if (unlikely(!put_literal(control, st, 0, 0)))
-		return false;
-	if (unlikely(!put_u32(control, st->ss, st->cksum)))
-		return false;
-	return true;
+	put_literal(control, st, 0, 0);
+	put_u32(control, st->ss, st->cksum);
 }
 
 
@@ -784,7 +774,6 @@ static inline void init_hash_indexes(struct rzip_state *st)
 
 #if defined(__APPLE__) || defined(__FreeBSD__)
 # define mremap fake_mremap
-#endif
 
 static inline void *fake_mremap(void *old_address, size_t old_size, size_t new_size, int flags __UNUSED__)
 {
@@ -804,12 +793,13 @@ static inline void *fake_mremap(void *old_address, size_t old_size, size_t new_s
 		return old_address;
 	}
 }
+#endif
 
 /* stdin is not file backed so we have to emulate the mmap by mapping
  * anonymous ram and reading stdin into it. It means the maximum ram
  * we can use will be less but we will already have determined this in
  * rzip_chunk */
-static inline bool mmap_stdin(rzip_control *control, uchar *buf,
+static inline void mmap_stdin(rzip_control *control, uchar *buf,
 			      struct rzip_state *st)
 {
 	i64 len = st->chunk_size;
@@ -822,7 +812,7 @@ static inline bool mmap_stdin(rzip_control *control, uchar *buf,
 		ret = MIN(len, one_g);
 		ret = read(fileno(control->inFILE), offset_buf, (size_t)ret);
 		if (unlikely(ret < 0))
-			fatal_return(("Failed to read in mmap_stdin\n"), false);
+			failure("Failed to read in mmap_stdin\n");
 		total += ret;
 		if (ret == 0) {
 			/* Should be EOF */
@@ -837,7 +827,7 @@ static inline bool mmap_stdin(rzip_control *control, uchar *buf,
 				st->chunk_size = 0;
 			}
 			if (unlikely(buf == MAP_FAILED))
-				fatal_return(("Failed to remap to smaller buf in mmap_stdin\n"), false);
+				failure("Failed to remap to smaller buf in mmap_stdin\n");
 			control->eof = st->stdin_eof = 1;
 			break;
 		}
@@ -845,24 +835,20 @@ static inline bool mmap_stdin(rzip_control *control, uchar *buf,
 		len -= ret;
 	}
 	control->st_size += total;
-	return true;
 }
 
-static inline bool
+static inline void
 init_sliding_mmap(rzip_control *control, struct rzip_state *st, int fd_in,
 		  i64 offset)
 {
 	struct sliding_buffer *sb = &control->sb;
 
-	/* Initialise the high buffer */
+	/* Initialise the high buffer. One page size is fastest to manipulate */
 	if (!STDIN) {
-		sb->high_length = 65536;
-		/* Round up to the next biggest page size */
-		if (sb->high_length % control->page_size)
-			sb->high_length += control->page_size - (sb->high_length % control->page_size);
+		sb->high_length = control->page_size;
 		sb->buf_high = (uchar *)mmap(NULL, sb->high_length, PROT_READ, MAP_SHARED, fd_in, offset);
 		if (unlikely(sb->buf_high == MAP_FAILED))
-			fatal_return(("Unable to mmap buf_high in init_sliding_mmap\n"), false);
+			failure("Unable to mmap buf_high in init_sliding_mmap\n");
 		sb->size_high = sb->high_length;
 		sb->offset_high = 0;
 	}
@@ -871,49 +857,43 @@ init_sliding_mmap(rzip_control *control, struct rzip_state *st, int fd_in,
 	sb->size_low = st->mmap_size;
 	sb->orig_size = st->chunk_size;
 	sb->fd = fd_in;
-	return true;
 }
 
 /* compress a chunk of an open file. Assumes that the file is able to
    be mmap'd and is seekable */
-static inline bool
+static inline void
 rzip_chunk(rzip_control *control, struct rzip_state *st, int fd_in, int fd_out,
 	   i64 offset, double pct_base, double pct_multiple)
 {
 	struct sliding_buffer *sb = &control->sb;
 
-	if (unlikely(!init_sliding_mmap(control, st, fd_in, offset)))
-		return false;
+	init_sliding_mmap(control, st, fd_in, offset);
 
 	st->ss = open_stream_out(control, fd_out, NUM_STREAMS, st->chunk_size, st->chunk_bytes);
 	if (unlikely(!st->ss))
-		fatal_return(("Failed to open streams in rzip_chunk\n"), false);
+		failure("Failed to open streams in rzip_chunk\n");
 
 	print_verbose("Beginning rzip pre-processing phase\n");
-	if (unlikely(!hash_search(control, st, pct_base, pct_multiple))) {
-		close_stream_out(control, st->ss);
-		return false;
-	}
+	hash_search(control, st, pct_base, pct_multiple);
 
 	/* unmap buffer before closing and reallocating streams */
 	if (unlikely(munmap(sb->buf_low, sb->size_low))) {
 		close_stream_out(control, st->ss);
-		fatal_return(("Failed to munmap in rzip_chunk\n"), false);
+		failure("Failed to munmap in rzip_chunk\n");
 	}
 	if (!STDIN) {
 		if (unlikely(munmap(sb->buf_high, sb->size_high))) {
 			close_stream_out(control, st->ss);
-			fatal_return(("Failed to munmap in rzip_chunk\n"), false);
+			failure("Failed to munmap in rzip_chunk\n");
 		}
 	}
 
 	if (unlikely(close_stream_out(control, st->ss)))
-		fatal_return(("Failed to flush/close streams in rzip_chunk\n"), false);
-	return true;
+		failure("Failed to flush/close streams in rzip_chunk\n");
 }
 
 /* compress a whole file chunks at a time */
-bool rzip_fd(rzip_control *control, int fd_in, int fd_out)
+void rzip_fd(rzip_control *control, int fd_in, int fd_out)
 {
 	struct sliding_buffer *sb = &control->sb;
 
@@ -926,31 +906,32 @@ bool rzip_fd(rzip_control *control, int fd_in, int fd_out)
 	struct timeval current, start, last;
 	i64 len = 0, last_chunk = 0;
 	int pass = 0, passes, j;
+	double chunkmbs, tdiff;
 	struct rzip_state *st;
 	struct statvfs fbuf;
 	struct stat s, s2;
-	double chunkmbs;
 	i64 free_space;
 
 	init_mutex(control, &control->control_lock);
 	if (!NO_MD5)
 		md5_init_ctx(&control->ctx);
-	init_mutex(control, &control->cksumlock);
+	cksem_init(control, &control->cksumsem);
+	cksem_post(control, &control->cksumsem);
 
 	st = calloc(sizeof(*st), 1);
 	if (unlikely(!st))
-		fatal_return(("Failed to allocate control state in rzip_fd\n"), false);
+		failure("Failed to allocate control state in rzip_fd\n");
 
 	if (LZO_COMPRESS) {
 		if (unlikely(lzo_init() != LZO_E_OK)) {
 			free(st);
-			fatal_return(("lzo_init() failed\n"), false);
+			failure("lzo_init() failed\n");
 		}
 	}
 
 	if (unlikely(fstat(fd_in, &s))) {
 		free(st);
-		fatal_return(("Failed to stat fd_in in rzip_fd\n"), false);
+		failure("Failed to stat fd_in in rzip_fd\n");
 	}
 
 	if (!STDIN) {
@@ -965,7 +946,7 @@ bool rzip_fd(rzip_control *control, int fd_in, int fd_out)
 		* uncompressed file. */
 		if (unlikely(fstatvfs(fd_out, &fbuf))) {
 			free(st);
-			fatal_return(("Failed to fstatvfs in compress_file\n"), false);
+			failure("Failed to fstatvfs in compress_file\n");
 		}
 		free_space = (i64)fbuf.f_bsize * (i64)fbuf.f_bavail;
 		if (free_space < control->st_size) {
@@ -973,7 +954,7 @@ bool rzip_fd(rzip_control *control, int fd_in, int fd_out)
 				print_err("Warning, possibly inadequate free space detected, but attempting to compress due to -f option being used.\n");
 			else {
 				free(st);
-				failure_return(("Possibly inadequate free space to compress file, use -f to override.\n"), false);
+				failure("Possibly inadequate free space to compress file, use -f to override.\n");
 			}
 		}
 	}
@@ -1019,8 +1000,10 @@ bool rzip_fd(rzip_control *control, int fd_in, int fd_out)
 	gettimeofday(&start, NULL);
 
 	prepare_streamout_threads(control);
-	control->get_sb = single_get_sb;
 	control->do_mcpy = single_mcpy;
+	control->next_tag = &single_next_tag;
+	control->full_tag = &single_full_tag;
+	control->match_len = &single_match_len;
 
 	while (!pass || len > 0 || (STDIN && !st->stdin_eof)) {
 		double pct_base, pct_multiple;
@@ -1047,7 +1030,7 @@ retry:
 					close_streamout_threads(control);
 					free(st->hash_table);
 					free(st);
-					fatal_return(("Failed to mmap %s\n", control->infile), false);
+					failure("Failed to mmap %s\n", control->infile);
 				}
 				st->mmap_size = st->mmap_size / 10 * 9;
 				round_to_page(&st->mmap_size);
@@ -1055,17 +1038,12 @@ retry:
 					close_streamout_threads(control);
 					free(st->hash_table);
 					free(st);
-					fatal_return(("Unable to mmap any ram\n"), false);
+					failure("Unable to mmap any ram\n");
 				}
 				goto retry;
 			}
 			st->chunk_size = st->mmap_size;
-			if (unlikely(!mmap_stdin(control, sb->buf_low, st))) {
-				close_streamout_threads(control);
-				free(st->hash_table);
-				free(st);
-				return false;
-			}
+			mmap_stdin(control, sb->buf_low, st);
 		} else {
 			/* NOTE The buf is saved here for !STDIN mode */
 			sb->buf_low = (uchar *)mmap(sb->buf_low, st->mmap_size, PROT_READ, MAP_SHARED, fd_in, offset);
@@ -1074,7 +1052,7 @@ retry:
 					close_streamout_threads(control);
 					free(st->hash_table);
 					free(st);
-					fatal_return(("Failed to mmap %s\n", control->infile), false);
+					failure("Failed to mmap %s\n", control->infile);
 				}
 				st->mmap_size = st->mmap_size / 10 * 9;
 				round_to_page(&st->mmap_size);
@@ -1082,14 +1060,16 @@ retry:
 					close_streamout_threads(control);
 					free(st->hash_table);
 					free(st);
-					fatal_return(("Unable to mmap any ram\n"), false);
+					failure("Unable to mmap any ram\n");
 				}
 				goto retry;
 			}
 			if (st->mmap_size < st->chunk_size) {
 				print_maxverbose("Enabling sliding mmap mode and using mmap of %lld bytes with window of %lld bytes\n", st->mmap_size, st->chunk_size);
-				control->get_sb = &sliding_get_sb;
 				control->do_mcpy = &sliding_mcpy;
+				control->next_tag = &sliding_next_tag;
+				control->full_tag = &sliding_full_tag;
+				control->match_len = &sliding_match_len;
 			}
 		}
 		print_maxverbose("Succeeded in testing %lld sized mmap for rzip pre-processing\n", st->mmap_size);
@@ -1131,12 +1111,12 @@ retry:
 
 		gettimeofday(&current, NULL);
 		/* this will count only when size > window */
-		if (last.tv_sec > 0) {
+		if (last.tv_sec > 0 && pct_base > 100) {
 			unsigned int eta_hours, eta_minutes, eta_seconds, elapsed_time, finish_time,
 				elapsed_hours, elapsed_minutes, elapsed_seconds, diff_seconds;
 
 			elapsed_time = current.tv_sec - start.tv_sec;
-			finish_time = elapsed_time / ((pct_base / 100.0) ? : 1);
+			finish_time = elapsed_time / (pct_base / 100.0);
 			elapsed_hours = elapsed_time / 3600;
 			elapsed_minutes = (elapsed_time / 60) % 60;
 			elapsed_seconds = elapsed_time % 60;
@@ -1159,12 +1139,7 @@ retry:
 
 		if (st->chunk_size == len)
 			control->eof = 1;
-		if (unlikely(!rzip_chunk(control, st, fd_in, fd_out, offset, pct_base, pct_multiple))) {
-			close_streamout_threads(control);
-			free(st->hash_table);
-			free(st);
-			return false;
-		}
+		rzip_chunk(control, st, fd_in, fd_out, offset, pct_base, pct_multiple);
 
 		/* st->chunk_size may be shrunk in rzip_chunk */
 		last_chunk = st->chunk_size;
@@ -1173,7 +1148,7 @@ retry:
 			close_streamout_threads(control);
 			free(st->hash_table);
 			free(st);
-			failure_return(("Wrote EOF to file yet chunk_size was shrunk, corrupting archive.\n"), false);
+			failure("Wrote EOF to file yet chunk_size was shrunk, corrupting archive.\n");
 		}
 	}
 
@@ -1181,7 +1156,7 @@ retry:
 		free(st->hash_table);
 	if (unlikely(!close_streamout_threads(control))) {
 		free(st);
-		return false;
+		failure("Failed to close_streamout_threads in rzip_fd\n");
 	}
 
 	if (!NO_MD5) {
@@ -1197,24 +1172,28 @@ retry:
 		if (ENCRYPT)
 			if (unlikely(!lrz_encrypt(control, control->md5_resblock, MD5_DIGEST_SIZE, control->salt_pass))) {
 				free(st);
-				return false;
+				failure("Failed to lrz_encrypt in rzip_fd\n");
 			}
 		if (unlikely(write_1g(control, control->md5_resblock, MD5_DIGEST_SIZE) != MD5_DIGEST_SIZE)) {
 			free(st);
-			fatal_return(("Failed to write md5 in rzip_fd\n"), false);
+			failure("Failed to write md5 in rzip_fd\n");
 		}
 	}
 
-	if (TMP_OUTBUF)
+	if (TMP_OUTBUF) {
 		if (unlikely(!flush_tmpoutbuf(control))) {
 			free(st);
-			return false;
+			failure("Failed to flush_tmpoutbuf in rzip_fd\n");
 		}
+	}
 
 	gettimeofday(&current, NULL);
 	if (STDIN)
 		s.st_size = control->st_size;
-	chunkmbs = (s.st_size / 1024 / 1024) / ((double)(current.tv_sec-start.tv_sec)? : 1);
+	tdiff = current.tv_sec - start.tv_sec;
+	if (!tdiff)
+		tdiff = 1;
+	chunkmbs = (s.st_size / 1024 / 1024) / tdiff;
 
 	fstat(fd_out, &s2);
 
@@ -1234,7 +1213,6 @@ retry:
 		       1.0 * s.st_size / s2.st_size, chunkmbs);
 
 	free(st);
-	return true;
 }
 
 void rzip_control_free(rzip_control *control)
