@@ -1,6 +1,6 @@
 /*
    Copyright (C) 2011 Serge Belyshev
-   Copyright (C) 2006-2013 Con Kolivas
+   Copyright (C) 2006-2015 Con Kolivas
    Copyright (C) 2011 Peter Hyman
    Copyright (C) 1998 Andrew Tridgell
 
@@ -67,7 +67,7 @@ static struct compress_thread{
 	uchar c_type;	/* Compression type */
 	i64 s_len;	/* Data length uncompressed */
 	i64 c_len;	/* Data length compressed */
-	pthread_mutex_t mutex; /* This thread's mutex */
+	cksem_t cksem;  /* This thread's semaphore */
 	struct stream_info *sinfo;
 	int streamno;
 	uchar salt[SALT_LEN];
@@ -113,17 +113,6 @@ bool lock_mutex(rzip_control *control, pthread_mutex_t *mutex)
 	return true;
 }
 
-/* Lock and unlock a mutex */
-bool wait_mutex(rzip_control *control, pthread_mutex_t *mutex)
-{
-	bool ret;
-
-	ret = lock_mutex(control, mutex);
-	if (likely(ret))
-		ret = unlock_mutex(control, mutex);
-	return ret;
-}
-
 static bool cond_wait(rzip_control *control, pthread_cond_t *cond, pthread_mutex_t *mutex)
 {
 	if (unlikely(pthread_cond_wait(cond, mutex)))
@@ -177,13 +166,14 @@ static int lzo_compresses(rzip_control *control, uchar *s_buf, i64 s_len);
 
 static int zpaq_compress_buf(rzip_control *control, struct compress_thread *cthread, long thread)
 {
+	i64 c_len, c_size;
 	uchar *c_buf;
-	i64 c_len;
 
 	if (!lzo_compresses(control, cthread->s_buf, cthread->s_len))
 		return 0;
 
-	c_buf = malloc(cthread->s_len + control->page_size);
+	c_size = round_up_page(control, cthread->s_len + 10000);
+	c_buf = malloc(c_size);
 	if (!c_buf) {
 		print_err("Unable to allocate c_buf in zpaq_compress_buf\n");
 		return -1;
@@ -210,7 +200,7 @@ static int zpaq_compress_buf(rzip_control *control, struct compress_thread *cthr
 
 static int bzip2_compress_buf(rzip_control *control, struct compress_thread *cthread)
 {
-	u32 dlen = cthread->s_len;
+	u32 dlen = round_up_page(control, cthread->s_len);
 	int bzip2_ret;
 	uchar *c_buf;
 
@@ -259,7 +249,7 @@ static int bzip2_compress_buf(rzip_control *control, struct compress_thread *cth
 
 static int gzip_compress_buf(rzip_control *control, struct compress_thread *cthread)
 {
-	unsigned long dlen = cthread->s_len;
+	unsigned long dlen = round_up_page(control, cthread->s_len);
 	uchar *c_buf;
 	int gzip_ret;
 
@@ -314,11 +304,13 @@ static int lzma_compress_buf(rzip_control *control, struct compress_thread *cthr
 		return 0;
 
 	/* only 7 levels with lzma, scale them */
-	lzma_level = control->compression_level * 7 / 9 ? : 1;
+	lzma_level = control->compression_level * 7 / 9;
+	if (!lzma_level)
+		lzma_level = 1;
 
 	print_maxverbose("Starting lzma back end compression thread...\n");
 retry:
-	dlen = cthread->s_len;
+	dlen = round_up_page(control, cthread->s_len);
 	c_buf = malloc(dlen);
 	if (!c_buf) {
 		print_err("Unable to allocate c_buf in lzma_compress_buf\n");
@@ -398,7 +390,7 @@ retry:
 static int lzo_compress_buf(rzip_control *control, struct compress_thread *cthread)
 {
 	lzo_uint in_len = cthread->s_len;
-	lzo_uint dlen = in_len + in_len / 16 + 64 + 3;
+	lzo_uint dlen = round_up_page(control, in_len + in_len / 16 + 64 + 3);
 	lzo_bytep wrkmem;
 	uchar *c_buf;
 	int ret = -1;
@@ -450,7 +442,7 @@ static int zpaq_decompress_buf(rzip_control *control __UNUSED__, struct uncomp_t
 	int ret = 0;
 
 	c_buf = ucthread->s_buf;
-	ucthread->s_buf = malloc(dlen);
+	ucthread->s_buf = malloc(round_up_page(control, dlen));
 	if (unlikely(!ucthread->s_buf)) {
 		print_err("Failed to allocate %ld bytes for decompression\n", dlen);
 		ret = -1;
@@ -480,7 +472,7 @@ static int bzip2_decompress_buf(rzip_control *control __UNUSED__, struct uncomp_
 	uchar *c_buf;
 
 	c_buf = ucthread->s_buf;
-	ucthread->s_buf = malloc(dlen);
+	ucthread->s_buf = malloc(round_up_page(control, dlen));
 	if (unlikely(!ucthread->s_buf)) {
 		print_err("Failed to allocate %d bytes for decompression\n", dlen);
 		ret = -1;
@@ -515,7 +507,7 @@ static int gzip_decompress_buf(rzip_control *control __UNUSED__, struct uncomp_t
 	uchar *c_buf;
 
 	c_buf = ucthread->s_buf;
-	ucthread->s_buf = malloc(dlen);
+	ucthread->s_buf = malloc(round_up_page(control, dlen));
 	if (unlikely(!ucthread->s_buf)) {
 		print_err("Failed to allocate %ld bytes for decompression\n", dlen);
 		ret = -1;
@@ -545,13 +537,13 @@ out:
 
 static int lzma_decompress_buf(rzip_control *control, struct uncomp_thread *ucthread)
 {
-	size_t dlen = (size_t)ucthread->u_len;
+	size_t dlen = ucthread->u_len;
 	int ret = 0, lzmaerr;
 	uchar *c_buf;
 	SizeT c_len = ucthread->c_len;
 
 	c_buf = ucthread->s_buf;
-	ucthread->s_buf = malloc(dlen);
+	ucthread->s_buf = malloc(round_up_page(control, dlen));
 	if (unlikely(!ucthread->s_buf)) {
 		print_err("Failed to allocate %lld bytes for decompression\n", (i64)dlen);
 		ret = -1;
@@ -588,7 +580,7 @@ static int lzo_decompress_buf(rzip_control *control __UNUSED__, struct uncomp_th
 	uchar *c_buf;
 
 	c_buf = ucthread->s_buf;
-	ucthread->s_buf = malloc(dlen);
+	ucthread->s_buf = malloc(round_up_page(control, dlen));
 	if (unlikely(!ucthread->s_buf)) {
 		print_err("Failed to allocate %lu bytes for decompression\n", (unsigned long)dlen);
 		ret = -1;
@@ -629,11 +621,17 @@ ssize_t put_fdout(rzip_control *control, void *offset_buf, ssize_t ret)
 		/* The data won't fit in a temporary output buffer so we have
 		 * to fall back to temporary files. */
 		print_verbose("Unable to decompress entirely in ram, will use physical files\n");
-		if (unlikely(!write_fdout(control, control->tmp_outbuf, control->out_len)))
+		if (unlikely(control->fd_out == -1))
+			failure("Was unable to decompress entirely in ram and no temporary file creation was possible\n");
+		if (unlikely(!write_fdout(control, control->tmp_outbuf, control->out_len))) {
+			print_err("Unable to write_fdout tmpoutbuf in put_fdout\n");
 			return -1;
+		}
 		close_tmpoutbuf(control);
-		if (unlikely(!write_fdout(control, offset_buf, ret)))
+		if (unlikely(!write_fdout(control, offset_buf, ret))) {
+			print_err("Unable to write_fdout offset_buf in put_fdout\n");
 			return -1;
+		}
 		return ret;
 	}
 	memcpy(control->tmp_outbuf + control->out_ofs, offset_buf, ret);
@@ -665,6 +663,8 @@ ssize_t write_1g(rzip_control *control, void *buf, i64 len)
 	return total;
 }
 
+/* Should be called only if we know the buffer will be large enough, otherwise
+ * we must dump_stdin first */
 static bool read_fdin(struct rzip_control *control, i64 len)
 {
 	int tmpchar;
@@ -681,6 +681,17 @@ static bool read_fdin(struct rzip_control *control, i64 len)
 	return true;
 }
 
+/* Dump STDIN into a temporary file */
+static int dump_stdin(rzip_control *control)
+{
+	if (unlikely(!write_fdin(control)))
+		return -1;
+	if (unlikely(!read_tmpinfile(control, control->fd_in)))
+		return -1;
+	close_tmpinbuf(control);
+	return 0;
+}
+
 /* Ditto for read */
 ssize_t read_1g(rzip_control *control, int fd, void *buf, i64 len)
 {
@@ -692,16 +703,13 @@ ssize_t read_1g(rzip_control *control, int fd, void *buf, i64 len)
 		/* We're decompressing from STDIN */
 		if (unlikely(control->in_ofs + len > control->in_maxlen)) {
 			/* We're unable to fit it all into the temp buffer */
-			if (unlikely(!write_fdin(control)))
-				return -1;
-			if (unlikely(!read_tmpinfile(control, control->fd_in)))
-				return -1;
-			close_tmpinbuf(control);
+			dump_stdin(control);
 			goto read_fd;
 		}
-		if (control->in_ofs + len > control->in_len)
+		if (control->in_ofs + len > control->in_len) {
 			if (unlikely(!read_fdin(control, control->in_ofs + len - control->in_len)))
 				return false;
+		}
 		memcpy(buf, control->tmp_inbuf + control->in_ofs, len);
 		control->in_ofs += len;
 		return len;
@@ -787,14 +795,6 @@ static inline int read_u32(rzip_control *control, int f, u32 *v)
 	return ret;
 }
 
-static inline int read_i64(rzip_control *control, int f, i64 *v)
-{
-	int ret = read_buf(control, f, (uchar *)v, 8);
-
-	*v = le64toh(*v);
-	return ret;
-}
-
 static inline int read_val(rzip_control *control, int f, i64 *v, int len)
 {
 	int ret;
@@ -837,9 +837,18 @@ static int read_seekto(rzip_control *control, struct stream_info *sinfo, i64 pos
 	i64 spos = pos + sinfo->initial_pos;
 
 	if (TMP_INBUF) {
-		if (spos > control->in_len)
-			if (unlikely(!read_fdin(control, spos - control->in_len)))
-				return -1;
+		if (spos > control->in_len) {
+			i64 len = spos - control->in_len;
+
+			if (control->in_ofs + len > control->in_maxlen) {
+				if (unlikely(dump_stdin(control)))
+					return -1;
+				goto fd_seek;
+			} else {
+				if (unlikely(!read_fdin(control, len)))
+					return -1;
+			}
+		}
 		control->in_ofs = spos;
 		if (unlikely(spos < 0)) {
 			print_err("Trying to seek to %lld outside tmp inbuf in read_seekto\n", spos);
@@ -847,7 +856,7 @@ static int read_seekto(rzip_control *control, struct stream_info *sinfo, i64 pos
 		}
 		return 0;
 	}
-
+fd_seek:
 	return fd_seekto(control, sinfo, spos, pos);
 }
 
@@ -897,14 +906,10 @@ bool prepare_streamout_threads(rzip_control *control)
 		fatal_return(("Unable to calloc cthread in prepare_streamout_threads\n"), false);
 	}
 
-	for (i = 0; i < control->threads; i++)
-		if (unlikely(!init_mutex(control, &cthread[i].mutex))) {
-			int x;
-			for (x = 0; x < i; x++) pthread_mutex_destroy(&cthread[x].mutex);
-			free(threads);
-			free(cthread);
-			return false;
-		}
+	for (i = 0; i < control->threads; i++) {
+		cksem_init(control, &cthread[i].cksem);
+		cksem_post(control, &cthread[i].cksem);
+	}
 	return true;
 }
 
@@ -916,13 +921,7 @@ bool close_streamout_threads(rzip_control *control)
 	/* Wait for the threads in the correct order in case they end up
 	 * serialised */
 	for (i = 0; i < control->threads; i++) {
-		if (unlikely(!lock_mutex(control, &cthread[close_thread].mutex))) {
-			int x;
-			for (x = 0; x < i; x++) unlock_mutex(control, &cthread[close_thread].mutex);
-			free(cthread);
-			free(threads);
-			return false;
-		}
+		cksem_wait(control, &cthread[close_thread].cksem);
 
 		if (++close_thread == control->threads)
 			close_thread = 0;
@@ -1343,8 +1342,10 @@ retry:
 				write_magic(control);
 			unlock_mutex(control, &control->control_lock);
 
-			if (unlikely(!flush_tmpoutbuf(control)))
+			if (unlikely(!flush_tmpoutbuf(control))) {
+				print_err("Failed to flush_tmpoutbuf in compthread\n");
 				goto error;
+			}
 		}
 
 		print_maxverbose("Writing initial chunk bytes value %d at %lld\n",
@@ -1439,18 +1440,18 @@ retry:
 	unlock_mutex(control, &output_lock);
 
 error:
-	unlock_mutex(control, &cti->mutex);
+	cksem_post(control, &cti->cksem);
 
 	return NULL;
 }
 
-static bool clear_buffer(rzip_control *control, struct stream_info *sinfo, int streamno, int newbuf)
+static void clear_buffer(rzip_control *control, struct stream_info *sinfo, int streamno, int newbuf)
 {
-	static long i = 0;
 	stream_thread_struct *s;
+	static int i = 0;
 
 	/* Make sure this thread doesn't already exist */
-	lock_mutex(control, &cthread[i].mutex);
+	cksem_wait(control, &cthread[i].cksem);
 
 	cthread[i].sinfo = sinfo;
 	cthread[i].streamno = streamno;
@@ -1462,38 +1463,32 @@ static bool clear_buffer(rzip_control *control, struct stream_info *sinfo, int s
 
 	s = malloc(sizeof(stream_thread_struct));
 	if (unlikely(!s)) {
-		unlock_mutex(control, &cthread[i].mutex);
-		fatal_return(("Unable to malloc in clear_buffer"), false);
+		cksem_post(control, &cthread[i].cksem);
+		failure("Unable to malloc in clear_buffer");
 	}
 	s->i = i;
 	s->control = control;
 	if (unlikely((!create_pthread(control, &threads[i], NULL, compthread, s)) ||
-	             (!detach_pthread(control, &threads[i])))) {
-		unlock_mutex(control, &cthread[i].mutex);
-		return false;
-	}
-
+	             (!detach_pthread(control, &threads[i]))))
+		failure("Unable to create compthread in clear_buffer");
 
 	if (newbuf) {
 		/* The stream buffer has been given to the thread, allocate a
 		 * new one. */
 		sinfo->s[streamno].buf = malloc(sinfo->bufsize);
-		if (unlikely(!sinfo->s[streamno].buf)) {
-			unlock_mutex(control, &cthread[i].mutex);
-			fatal_return(("Unable to malloc buffer of size %lld in flush_buffer\n", sinfo->bufsize), false);
-		}
+		if (unlikely(!sinfo->s[streamno].buf))
+			failure("Unable to malloc buffer of size %lld in flush_buffer\n", sinfo->bufsize);
 		sinfo->s[streamno].buflen = 0;
 	}
 
 	if (++i == control->threads)
 		i = 0;
-	return true;
 }
 
 /* flush out any data in a stream buffer */
-bool flush_buffer(rzip_control *control, struct stream_info *sinfo, int streamno)
+void flush_buffer(rzip_control *control, struct stream_info *sinfo, int streamno)
 {
-	return clear_buffer(control, sinfo, streamno, 1);
+	clear_buffer(control, sinfo, streamno, 1);
 }
 
 static void *ucompthread(void *data)
@@ -1612,6 +1607,7 @@ fill_another:
 			return -1;
 		header_length = 1 + (read_len * 3);
 	}
+	print_maxverbose("Fill_buffer stream %d c_len %lld u_len %lld last_head %lld\n", streamno, c_len, u_len, last_head);
 	sinfo->total_read += header_length;
 
 	if (ENCRYPT) {
@@ -1698,7 +1694,7 @@ out:
 }
 
 /* write some data to a stream. Return -1 on failure */
-int write_stream(rzip_control *control, void *ss, int streamno, uchar *p, i64 len)
+void write_stream(rzip_control *control, void *ss, int streamno, uchar *p, i64 len)
 {
 	struct stream_info *sinfo = ss;
 
@@ -1714,10 +1710,8 @@ int write_stream(rzip_control *control, void *ss, int streamno, uchar *p, i64 le
 
 		/* Flush the buffer every sinfo->bufsize into one thread */
 		if (sinfo->s[streamno].buflen == sinfo->bufsize)
-			if (unlikely(!flush_buffer(control, sinfo, streamno)))
-				return -1;
+			flush_buffer(control, sinfo, streamno);
 	}
-	return 0;
 }
 
 /* read some data from a stream. Return number of bytes read, or -1
@@ -1757,10 +1751,8 @@ int close_stream_out(rzip_control *control, void *ss)
 	struct stream_info *sinfo = ss;
 	int i;
 
-	for (i = 0; i < sinfo->num_streams; i++) {
-		if (unlikely(!clear_buffer(control, sinfo, i, 0)))
-			return -1;
-	}
+	for (i = 0; i < sinfo->num_streams; i++)
+		clear_buffer(control, sinfo, i, 0);
 
 	if (ENCRYPT) {
 		/* Last two compressed blocks do not have an offset written
@@ -1769,8 +1761,8 @@ int close_stream_out(rzip_control *control, void *ss)
 		int close_thread = output_thread;
 
 		for (i = 0; i < control->threads; i++) {
-			lock_mutex(control, &cthread[close_thread].mutex);
-			unlock_mutex(control, &cthread[close_thread].mutex);
+			cksem_wait(control, &cthread[close_thread].cksem);
+			cksem_post(control, &cthread[close_thread].cksem);
 			if (++close_thread == control->threads)
 				close_thread = 0;
 		}
@@ -1781,16 +1773,20 @@ int close_stream_out(rzip_control *control, void *ss)
 		if (!control->sinfo_buckets) {
 			/* no streams added */
 			control->sinfo_queue = calloc(STREAM_BUCKET_SIZE + 1, sizeof(void*));
-			if (!control->sinfo_queue)
+			if (!control->sinfo_queue) {
+				print_err("Failed to calloc sinfo_queue in close_stream_out\n");
 				return -1;
+			}
 			control->sinfo_buckets++;
 		} else if (control->sinfo_idx == STREAM_BUCKET_SIZE * control->sinfo_buckets + 1) {
 			/* all buckets full, create new bucket */
 			void *tmp;
 
 			tmp = realloc(control->sinfo_queue, (++control->sinfo_buckets * STREAM_BUCKET_SIZE + 1) * sizeof(void*));
-			if (!tmp)
+			if (!tmp) {
+				print_err("Failed to realloc sinfo_queue in close_stream_out\n");
 				return -1;
+			}
 			control->sinfo_queue = tmp;
 			memset(control->sinfo_queue + control->sinfo_idx, 0, ((control->sinfo_buckets * STREAM_BUCKET_SIZE + 1) - control->sinfo_idx) * sizeof(void*));
 		}
