@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2006-2013 Con Kolivas
+   Copyright (C) 2006-2015 Con Kolivas
    Copyright (C) 2011 Serge Belyshev
    Copyright (C) 2008, 2011 Peter Hyman
    Copyright (C) 1998 Andrew Tridgell
@@ -110,7 +110,10 @@ void setup_overhead(rzip_control *control)
 	/* Work out the compression overhead per compression thread for the
 	 * compression back-ends that need a lot of ram */
 	if (LZMA_COMPRESS) {
-		int level = control->compression_level * 7 / 9 ? : 1;
+		int level = control->compression_level * 7 / 9;
+
+		if (!level)
+			level = 1;
 		i64 dictsize = (level <= 5 ? (1 << (level * 2 + 14)) :
 				(level == 6 ? (1 << 25) : (1 << 26)));
 
@@ -149,6 +152,15 @@ void round_to_page(i64 *size)
 		*size = PAGE_SIZE;
 }
 
+size_t round_up_page(rzip_control *control, size_t len)
+{
+	int rem = len % control->page_size;
+
+	if (rem)
+		len += control->page_size - rem;
+	return len;
+}
+
 bool get_rand(rzip_control *control, uchar *buf, int len)
 {
 	int fd, i;
@@ -164,7 +176,9 @@ bool get_rand(rzip_control *control, uchar *buf, int len)
 			fatal_return(("Failed to close fd in get_rand\n"), false);
 	}
 	return true;
-}bool read_config(rzip_control *control)
+}
+
+bool read_config(rzip_control *control)
 {
 	/* check for lrzip.conf in ., $HOME/.lrzip and /etc/lrzip */
 	char *HOME, homeconf[255];
@@ -176,11 +190,6 @@ bool get_rand(rzip_control *control, uchar *buf, int len)
 	fp = fopen("lrzip.conf", "r");
 	if (fp)
 		fprintf(control->msgout, "Using configuration file ./lrzip.conf\n");
-	if (fp == NULL) {
-		fp = fopen("/etc/lrzip/lrzip.conf", "r");
-		if (fp)
-			fprintf(control->msgout, "Using configuration file /etc/lrzip/lrzip.conf\n");
-	}
 	if (fp == NULL) {
 		HOME=getenv("HOME");
 		if (HOME) {
@@ -190,6 +199,11 @@ bool get_rand(rzip_control *control, uchar *buf, int len)
 				fprintf(control->msgout, "Using configuration file %s\n", homeconf);
 		}
 	}
+	if (fp == NULL) {
+		fp = fopen("/etc/lrzip/lrzip.conf", "r");
+		if (fp)
+			fprintf(control->msgout, "Using configuration file /etc/lrzip/lrzip.conf\n");
+	}
 	if (fp == NULL)
 		return true;
 
